@@ -2,7 +2,7 @@ use crate::{
     connection::Connection,
     ffi_util::IntoCString,
     macros::{Macros, Stage},
-    message::replies::Reply,
+    message::reply::Reply,
     proto_util::{Actions, ProtoOpts},
 };
 use async_trait::async_trait;
@@ -31,10 +31,13 @@ pub trait SetErrorReply {
 }
 
 /// An error that occurs when configuring the SMTP reply.
-#[derive(Debug)]
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
 pub enum SmtpReplyError {
+    /// The SMTP reply code is invalid.
     InvalidReplyCode,
+    /// The enhanced status code is invalid.
     InvalidEnhancedStatusCode,
+    /// The SMTP reply text is invalid.
     InvalidReplyText,
 }
 
@@ -54,87 +57,116 @@ impl Error for SmtpReplyError {}
 #[async_trait]
 pub trait ContextActions {
     /// Adds a header.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::ADD_HEADER`][crate::Actions::ADD_HEADER].
     async fn add_header<'cx, 'k, 'v>(
         &'cx self,
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Inserts a header at the given index.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::ADD_HEADER`][crate::Actions::ADD_HEADER].
     async fn insert_header<'cx, 'k, 'v>(
         &'cx self,
         index: i32,
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Replaces or deletes the header at the given index.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::CHANGE_HEADER`][crate::Actions::CHANGE_HEADER].
     async fn change_header<'cx, 'k, 'v>(
         &'cx self,
         name: impl IntoCString + Send + 'k,
         index: i32,
         value: Option<impl IntoCString + Send + 'v>,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Replaces the envelope sender, with optional ESMTP arguments.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::CHANGE_SENDER`][crate::Actions::CHANGE_SENDER].
     async fn change_sender<'cx, 'a, 'b>(
         &'cx self,
         mail: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Adds an envelope recipient.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::ADD_RCPT`][crate::Actions::ADD_RCPT].
     async fn add_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Adds an envelope recipient, with ESMTP arguments.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::ADD_RCPT_EXT`][crate::Actions::ADD_RCPT_EXT].
     async fn add_recipient_ext<'cx, 'a, 'b>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Removes an envelope recipient.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::DELETE_RCPT`][crate::Actions::DELETE_RCPT].
     async fn delete_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 
     /// Replaces the message body.
-    async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ContextError>;
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::REPLACE_BODY`][crate::Actions::REPLACE_BODY].
+    async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ActionError>;
 
     /// Reports operation-in-progress to the MTA.
-    async fn progress<'cx>(&'cx self) -> Result<(), ContextError>;
+    async fn progress<'cx>(&'cx self) -> Result<(), ActionError>;
 
     /// Quarantines the message.
+    ///
+    /// Must be enabled with flag
+    /// [`Actions::QUARANTINE`][crate::Actions::QUARANTINE].
     async fn quarantine<'cx, 'a>(
         &'cx self,
         reason: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError>;
+    ) -> Result<(), ActionError>;
 }
 
-// TODO Rename ActionError?
 /// An error that occurs when executing a context action.
 #[derive(Debug)]
-pub enum ContextError {
-    ActionUnavailable,
-    InvalidParams,
+pub enum ActionError {
+    /// The action has not been enabled.
+    NotAvailable,
+    /// A parameter given for the action is invalid.
+    InvalidParam,
+    /// An I/O error occurred.
     Io(io::Error),
 }
 
-impl Display for ContextError {
+impl Display for ActionError {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         match self {
-            Self::ActionUnavailable => write!(f, "action has not been enabled"),
-            Self::InvalidParams => write!(f, "invalid context action parameters"),
+            Self::NotAvailable => write!(f, "action has not been enabled"),
+            Self::InvalidParam => write!(f, "invalid context action parameter"),
             Self::Io(error) => write!(f, "I/O error: {}", error),
         }
     }
 }
 
-impl Error for ContextError {
+impl Error for ActionError {
     fn source(&self) -> Option<&(dyn Error + 'static)> {
         match self {
             Self::Io(error) => Some(error),
@@ -143,7 +175,7 @@ impl Error for ContextError {
     }
 }
 
-impl From<io::Error> for ContextError {
+impl From<io::Error> for ActionError {
     fn from(error: io::Error) -> Self {
         Self::Io(error)
     }
@@ -155,11 +187,45 @@ pub struct NegotiateContext<T: Send> {
     pub data: Option<T>,
     /// The custom SMTP reply.
     pub reply: SmtpReply,
+
     /// The actions to request from the MTA.
+    ///
+    /// Requested actions can also be specified in
+    /// [`Config::actions`][crate::Config::actions].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use indymilter::{Actions, NegotiateContext};
+    /// # fn f(context: &mut NegotiateContext<()>) {
+    /// context.requested_actions |= Actions::ADD_RCPT | Actions::ADD_RCPT_EXT;
+    /// # }
+    /// ```
     pub requested_actions: Actions,
+
     /// The milter protocol options to request from the MTA.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use indymilter::{NegotiateContext, ProtoOpts};
+    /// # fn f(context: &mut NegotiateContext<()>) {
+    /// context.requested_opts |= ProtoOpts::SKIP;
+    /// # }
+    /// ```
     pub requested_opts: ProtoOpts,
+
     /// The macros to request from the MTA.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use ::{indymilter::{NegotiateContext, Stage}, std::ffi::CString};
+    /// # fn f(context: &mut NegotiateContext<()>) {
+    /// let macros = CString::new("{auth_authen} {auth_type}").unwrap();
+    /// context.requested_macros.insert(Stage::Mail, macros);
+    /// # }
+    /// ```
     pub requested_macros: HashMap<Stage, CString>,
 }
 
@@ -227,7 +293,7 @@ impl<T: Send> Context<T> {
         F: FnMut(&ReplyCode) -> bool,
     {
         self.reply
-            .take_reply_if_match(predicate)
+            .take_error_reply_if(predicate)
             .map(|reply| Reply::ReplyCode { reply })
     }
 }
@@ -265,9 +331,10 @@ impl<T: Send> EomContext<T> {
 }
 
 // Note: This struct does *not* implement `Default` or `Clone`. Callbacks
-// receive a mutable reference to this struct but must not be able to remove it.
+// receive a mutable reference to this struct but must not be able to replace
+// it.
 /// Current custom SMTP reply.
-#[derive(Debug)]
+#[derive(Debug, Eq, Hash, PartialEq)]
 pub struct SmtpReply {
     error_reply: Option<ErrorReply>,
 }
@@ -277,14 +344,13 @@ impl SmtpReply {
         Self { error_reply: None }
     }
 
-    pub(crate) fn duplicate(&self) -> Self {
+    pub(crate) fn clone_internal(&self) -> Self {
         Self {
             error_reply: self.error_reply.clone(),
         }
     }
 
-    // TODO revisit
-    pub(crate) fn take_reply_if_match<F>(&mut self, mut predicate: F) -> Option<CString>
+    pub(crate) fn take_error_reply_if<F>(&mut self, mut predicate: F) -> Option<CString>
     where
         F: FnMut(&ReplyCode) -> bool,
     {
@@ -317,6 +383,9 @@ impl SetErrorReply for SmtpReply {
             .transpose()
             .map_err(|_| SmtpReplyError::InvalidEnhancedStatusCode)?;
 
+        // Like in libmilter, compatibility of reply code and enhanced status
+        // code is not checked.
+
         let mut message_lines = Vec::new();
 
         for (i, line) in message.into_iter().enumerate() {
@@ -346,7 +415,7 @@ impl SetErrorReply for SmtpReply {
     }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
 struct ErrorReply {
     rcode: ReplyCode,
     xcode: Option<EnhancedStatusCode>,
@@ -419,14 +488,14 @@ impl ContextActions for EomActions {
         &'cx self,
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::ADD_HEADER) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let value = value.into_c_string();
@@ -441,17 +510,17 @@ impl ContextActions for EomActions {
         index: i32,
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::ADD_HEADER) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
         if index < 0 {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let value = value.into_c_string();
@@ -466,17 +535,17 @@ impl ContextActions for EomActions {
         name: impl IntoCString + Send + 'k,
         index: i32,
         value: Option<impl IntoCString + Send + 'v>,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::CHANGE_HEADER) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
         if index < 0 {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let value = value.map_or_else(Default::default, |v| v.into_c_string());
@@ -490,14 +559,14 @@ impl ContextActions for EomActions {
         &'cx self,
         mail: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::CHANGE_SENDER) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let mail = mail.into_c_string();
         if mail.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let args = args.map(|a| a.into_c_string());
@@ -510,14 +579,14 @@ impl ContextActions for EomActions {
     async fn add_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::ADD_RCPT) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         self.conn.write_reply(Reply::AddRcpt { rcpt }).await?;
@@ -529,14 +598,14 @@ impl ContextActions for EomActions {
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::ADD_RCPT_EXT) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         let args = args.map(|a| a.into_c_string());
@@ -549,14 +618,14 @@ impl ContextActions for EomActions {
     async fn delete_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::DELETE_RCPT) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         self.conn.write_reply(Reply::DeleteRcpt { rcpt }).await?;
@@ -564,9 +633,9 @@ impl ContextActions for EomActions {
         Ok(())
     }
 
-    async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ContextError> {
+    async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::REPLACE_BODY) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         const CHUNK_SIZE: usize = 65535;
@@ -586,7 +655,7 @@ impl ContextActions for EomActions {
         Ok(())
     }
 
-    async fn progress<'cx>(&'cx self) -> Result<(), ContextError> {
+    async fn progress<'cx>(&'cx self) -> Result<(), ActionError> {
         self.conn.write_reply(Reply::Progress).await?;
 
         Ok(())
@@ -595,14 +664,14 @@ impl ContextActions for EomActions {
     async fn quarantine<'cx, 'a>(
         &'cx self,
         reason: impl IntoCString + Send + 'a,
-    ) -> Result<(), ContextError> {
+    ) -> Result<(), ActionError> {
         if !self.available_actions.contains(Actions::QUARANTINE) {
-            return Err(ContextError::ActionUnavailable);
+            return Err(ActionError::NotAvailable);
         }
 
         let reason = reason.into_c_string();
         if reason.as_bytes().is_empty() {
-            return Err(ContextError::InvalidParams);
+            return Err(ActionError::InvalidParam);
         }
 
         self.conn.write_reply(Reply::Quarantine { reason }).await?;
@@ -611,7 +680,7 @@ impl ContextActions for EomActions {
     }
 }
 
-#[derive(Clone, Copy, Debug, Default, Eq, Hash, PartialEq)]
+#[derive(Clone, Debug, Default, Eq, Hash, PartialEq)]
 pub struct ParseStatusCodeError;
 
 impl Error for ParseStatusCodeError {}
