@@ -1,14 +1,18 @@
 //! Milter protocol messages.
+//!
+//! This module contains low-level protocol helpers.
 
 use bytes::Bytes;
 use std::{
     ascii,
     error::Error,
     fmt::{self, Debug, Display, Formatter, Write},
+    io::{self, ErrorKind},
 };
+use tokio::io::{AsyncRead, AsyncReadExt, AsyncWrite, AsyncWriteExt};
 
-pub mod commands;
-pub mod replies;
+pub mod command;
+pub mod reply;
 
 /// The type of the milter protocol version.
 pub type Version = u32;
@@ -32,7 +36,7 @@ impl Message {
     //
     // Support for the experimental, negotiable custom `Maxdatasize` feature of
     // libmilter is not implemented.
-    pub(crate) const MAX_BUFFER_LEN: usize = 1024 * 1024 - 1;
+    const MAX_BUFFER_LEN: usize = 1024 * 1024 - 1;
 
     /// Creates a new message with the given kind and payload buffer.
     pub fn new(kind: impl Into<u8>, buffer: impl Into<Bytes>) -> Self {
@@ -52,7 +56,7 @@ impl Debug for Message {
     }
 }
 
-struct Byte(u8);
+pub(crate) struct Byte(pub u8);
 
 impl Debug for Byte {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
@@ -120,9 +124,54 @@ impl Display for TryFromIndexError {
 
 impl Error for TryFromIndexError {}
 
+/// Reads a milter protocol message from a stream.
+pub async fn read<S>(stream: &mut S) -> io::Result<Message>
+where
+    S: AsyncRead + Unpin,
+{
+    let len = stream.read_u32().await?;
+    let kind = stream.read_u8().await?;
+
+    let len = usize::try_from(len)
+        .expect("unsupported pointer size")
+        .saturating_sub(1);
+
+    if len > Message::MAX_BUFFER_LEN {
+        return Err(ErrorKind::InvalidData.into());
+    }
+
+    let mut buffer = vec![0; len];
+    stream.read_exact(&mut buffer).await?;
+
+    Ok(Message::new(kind, buffer))
+}
+
+/// Writes a milter protocol message to a stream.
+pub async fn write<S>(stream: &mut S, msg: Message) -> io::Result<()>
+where
+    S: AsyncWrite + Unpin,
+{
+    let len = msg.buffer.len();
+
+    if len > Message::MAX_BUFFER_LEN {
+        return Err(ErrorKind::InvalidData.into());
+    }
+
+    let len = u32::try_from(len).unwrap().checked_add(1).unwrap();
+
+    stream.write_u32(len).await?;
+    stream.write_u8(msg.kind).await?;
+    stream.write_all(msg.buffer.as_ref()).await?;
+    stream.flush().await?;
+
+    Ok(())
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
+    use std::io::ErrorKind;
+    use tokio::io::{AsyncReadExt, AsyncWriteExt};
 
     #[test]
     fn byte_debug_ok() {
@@ -154,4 +203,44 @@ mod tests {
             "failed to convert byte b'x'"
         );
     }
+
+    #[tokio::test]
+    async fn read_message() {
+        let mut stream = {
+            let (mut client, stream) = tokio::io::duplex(100);
+
+            client.write_all(b"\0\0\0\x02xyz").await.unwrap();
+
+            stream
+        };
+
+        let msg = read(&mut stream).await.unwrap();
+        assert_eq!(msg, Message::new(b'x', "y"));
+
+        let error = read(&mut stream).await.unwrap_err();
+        assert_eq!(error.kind(), ErrorKind::UnexpectedEof);
+    }
+
+    #[tokio::test]
+    async fn write_message() {
+        let mut client = {
+            let (client, mut stream) = tokio::io::duplex(100);
+
+            let msg = Message::new(b'x', "abc");
+            write(&mut stream, msg).await.unwrap();
+            stream.write_u8(1).await.unwrap();
+
+            client
+        };
+
+        let mut buffer = vec![0; 8];
+        client.read_exact(&mut buffer).await.unwrap();
+        assert_eq!(buffer, b"\0\0\0\x04xabc");
+
+        let byte = client.read_u8().await.unwrap();
+        assert_eq!(byte, 1);
+
+        let error = client.read_u8().await.unwrap_err();
+        assert_eq!(error.kind(), ErrorKind::UnexpectedEof);
+    }
 }
