@@ -5,12 +5,12 @@ use crate::{
     context::{Context, EomContext, NegotiateContext, ReplyCode},
     macros::Stage,
     message::{
-        commands::{
+        command::{
             CommandKind, CommandMessage, ConnInfoPayload, EnvAddrPayload, HeaderPayload,
             HeloPayload, MacroPayload, OptNegPayload, ParseCommandError, UnknownPayload,
         },
-        replies::Reply,
-        Version, PROTOCOL_VERSION,
+        reply::Reply,
+        Byte, Version, PROTOCOL_VERSION,
     },
     proto_util::{Actions, ProtoOpts},
 };
@@ -132,44 +132,53 @@ fn opts_from_callbacks<T: Send>(callbacks: &Callbacks<T>) -> ProtoOpts {
 
 type SessionResult = Result<(), SessionError>;
 
-/// Reasons for abnormal session termination.
+/// Causes of abnormal session termination.
 #[derive(Debug)]
 enum SessionError {
-    Io(io::Error),
-    ParseCommand(ParseCommandError),
-    ProtocolVersionNotSupported,
-    ActionsNotSupported,
-    ProtoOptsNotSupported,
-    InvalidNegotiateStatus(Status),
-    BufferEmpty(CommandKind),
-    UnknownCommand(u8),
     MilterShutDown,
+    Io(io::Error),
+    UnknownCommand(u8),
+    BufferEmpty,
+    ParseCommand(ParseCommandError),
+    ProtocolVersionNotSupported(Version),
+    InvalidNegotiateStatus(Status),
+    ActionsNotSupported(Actions),
+    ProtoOptsNotSupported(ProtoOpts),
 }
 
-// TODO human readable display for debugging
 impl Display for SessionError {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         match self {
+            Self::MilterShutDown => write!(f, "milter shut down"),
             Self::Io(error) => write!(f, "I/O error: {}", error),
+            Self::UnknownCommand(byte) => write!(f, "unknown command: {:?}", Byte(*byte)),
+            Self::BufferEmpty => write!(f, "command with empty payload buffer"),
             Self::ParseCommand(error) => write!(f, "command could not be parsed: {}", error),
-            Self::ProtocolVersionNotSupported => {
-                write!(f, "requested milter protocol version not supported")
-            }
-            Self::ActionsNotSupported => write!(f, "requested actions not supported"),
-            Self::ProtoOptsNotSupported => {
-                write!(f, "requested milter protocol options not supported")
+            Self::ProtocolVersionNotSupported(version) => {
+                write!(f, "requested milter protocol version {} not supported", version)
             }
             Self::InvalidNegotiateStatus(status) => {
                 write!(f, "invalid status in negotiation: {:?}", status)
             }
-            Self::BufferEmpty(cmd) => write!(f, "command with empty payload buffer: {:?}", cmd),
-            Self::UnknownCommand(byte) => write!(f, "unknown command: {:?}", byte),
-            Self::MilterShutDown => write!(f, "milter shut down"),
+            Self::ActionsNotSupported(actions) => {
+                write!(f, "requested actions not supported: {:?}", actions)
+            }
+            Self::ProtoOptsNotSupported(opts) => {
+                write!(f, "requested milter protocol options not supported: {:?}", opts)
+            }
         }
     }
 }
 
-impl Error for SessionError {}
+impl Error for SessionError {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
+        match self {
+            Self::Io(error) => Some(error),
+            Self::ParseCommand(error) => Some(error),
+            _ => None,
+        }
+    }
+}
 
 impl From<io::Error> for SessionError {
     fn from(error: io::Error) -> Self {
@@ -233,13 +242,9 @@ pub fn spawn<S, T>(
 
 struct Session<T: Send> {
     conn: Connection,
-
     state: State,
-
     shutdown: watch::Receiver<bool>,
-
     callbacks: Arc<Callbacks<T>>,
-
     context: Context<T>,
     actions: Actions,
     opts: ProtoOpts,
@@ -262,13 +267,9 @@ impl<T: Send> Session<T> {
 
         Self {
             conn,
-
             state: State::Init,
-
             shutdown,
-
             callbacks,
-
             context: Context::new(),
             actions,
             opts,
@@ -327,6 +328,8 @@ impl<T: Send> Session<T> {
 
             let cmd = msg.kind;
 
+            trace!(?cmd, "got next command");
+
             // Check if desired state change from the CommandMessage is feasible
             // taking into account the current set of protocol options.
 
@@ -460,6 +463,9 @@ impl<T: Send> Session<T> {
             }
             CommandKind::BodyChunk => {
                 if let Some(body) = &self.callbacks.body {
+                    // Convert `Bytes` to `Vec<u8>`. This allows milter writers
+                    // to not require a dependency on the `bytes` crate for
+                    // implementing a milter.
                     let chunk = msg.buffer.as_ref().into();
                     body(&mut self.context, chunk).await
                 } else {
@@ -488,8 +494,8 @@ impl<T: Send> Session<T> {
                         let mut cx = EomContext::new(
                             self.conn.clone(),
                             self.context.data.take(),
-                            self.context.macros.duplicate(),
-                            self.context.reply.duplicate(),
+                            self.context.macros.clone_internal(),
+                            self.context.reply.clone_internal(),
                             self.actions,
                         );
 
@@ -556,14 +562,14 @@ impl<T: Send> Session<T> {
         // Note: this is different from libmilter. We only support protocol
         // versions 6 (and up). Older versions are not supported.
         if mta_version < PROTOCOL_VERSION {
-            return Err(SessionError::ProtocolVersionNotSupported);
+            return Err(SessionError::ProtocolVersionNotSupported(mta_version));
         }
 
         if mta_actions.is_empty() {
-            mta_actions = Actions::min_actions();
+            mta_actions = Actions::min_flags();
         }
         if mta_opts.is_empty() {
-            mta_opts = ProtoOpts::min_proto_opts();
+            mta_opts = ProtoOpts::min_flags();
         }
 
         let target_opts;
@@ -575,7 +581,7 @@ impl<T: Send> Session<T> {
 
             let mut cx = NegotiateContext::new(
                 self.context.data.take(),
-                self.context.reply.duplicate(),
+                self.context.reply.clone_internal(),
                 default_actions,
                 default_opts,
             );
@@ -609,10 +615,10 @@ impl<T: Send> Session<T> {
         }
 
         if !mta_actions.contains(self.actions) {
-            return Err(SessionError::ActionsNotSupported);
+            return Err(SessionError::ActionsNotSupported(self.actions));
         }
         if !mta_opts.contains(target_opts) {
-            return Err(SessionError::ProtoOptsNotSupported);
+            return Err(SessionError::ProtoOptsNotSupported(target_opts));
         }
 
         self.write_opt_neg_reply(
@@ -733,7 +739,7 @@ fn ensure_buffer_present(cmd: CommandKind, buf: &[u8]) -> SessionResult {
         DefMacros | BodyChunk | ConnInfo | Helo | Header | Mail | OptNeg | Rcpt | Unknown
     ) && buf.is_empty()
     {
-        return Err(SessionError::BufferEmpty(cmd));
+        return Err(SessionError::BufferEmpty);
     }
 
     Ok(())
