@@ -5,79 +5,16 @@ use byte_strings::c_str;
 use bytes::Bytes;
 use indymilter::{
     message::{
-        commands::{Command, EnvAddrPayload, OptNegPayload},
-        replies::Reply,
+        command::{Command, EnvAddrPayload, OptNegPayload},
+        reply::Reply,
         PROTOCOL_VERSION,
     },
-    Actions, Callbacks, Config, ContextActions, ProtoOpts, SetErrorReply, Status,
+    ActionError, Actions, Callbacks, Config, ContextActions, ProtoOpts, SetErrorReply, Status,
 };
+use std::ffi::CString;
 
 #[tokio::test]
-async fn eom_basic() {
-    init_tracing_subscriber();
-
-    let callbacks = Callbacks::<()>::new()
-        .on_eom(|cx| {
-            Box::pin(async move {
-                cx.actions.add_header("name", "value").await.unwrap();
-
-                Status::Continue
-            })
-        });
-
-    let config = Config {
-        actions: Actions::ADD_HEADER,
-        ..Default::default()
-    };
-
-    let milter = Milter::spawn(LOCALHOST, callbacks, config).await.unwrap();
-
-    let mut client = Client::connect(milter.addr()).await.unwrap();
-
-    client
-        .write_command(Command::OptNeg(OptNegPayload {
-            version: PROTOCOL_VERSION,
-            actions: Actions::all(),
-            opts: ProtoOpts::all(),
-        }))
-        .await
-        .unwrap();
-
-    let reply = client.read_reply().await.unwrap();
-    assert!(matches!(
-        reply,
-        Reply::OptNeg {
-            actions: Actions::ADD_HEADER,
-            ..
-        }
-    ));
-
-    client
-        .write_command(Command::BodyEnd(Bytes::new()))
-        .await
-        .unwrap();
-
-    let reply = client.read_reply().await.unwrap();
-    assert_eq!(
-        reply,
-        Reply::AddHeader {
-            name: c_str!("name").into(),
-            value: c_str!("value").into(),
-        }
-    );
-
-    let reply = client.read_reply().await.unwrap();
-    assert_eq!(reply, Reply::Continue);
-
-    client.write_command(Command::Quit).await.unwrap();
-
-    client.disconnect().await.unwrap();
-
-    milter.shutdown().await.unwrap();
-}
-
-#[tokio::test]
-async fn reply_basic() {
+async fn custom_error_reply() {
     init_tracing_subscriber();
 
     let callbacks = Callbacks::<()>::new()
@@ -89,7 +26,7 @@ async fn reply_basic() {
             })
         });
 
-    let milter = Milter::spawn(LOCALHOST, callbacks, Default::default())
+    let milter = Milter::spawn(LOCALHOST, callbacks, default_config())
         .await
         .unwrap();
 
@@ -130,27 +67,29 @@ async fn reply_basic() {
 }
 
 #[tokio::test]
-async fn replace_body_chunks() {
+async fn modify_header_actions() {
     init_tracing_subscriber();
 
-    let new_body_size = 2_000_000;
-
     let callbacks = Callbacks::<()>::new()
-        .on_eom(move |cx| {
-            Box::pin(async move {
-                // New body with large size, greater than chunk size
-                // (implementation detail).
-                let body = vec![b'x'; new_body_size];
+        .on_eom(|cx| {
+            Box::pin(async {
+                // Also try passing a non-static string to `add_header` (to test
+                // `IntoCString + Send` parameter type).
+                let value = format!("<{}{}>", "value", 42);
 
-                cx.actions.replace_body(&body).await.unwrap();
+                cx.actions.add_header("name", value.as_str()).await.unwrap();
+                cx.actions
+                    .change_header("other", 1, Some("some value"))
+                    .await
+                    .unwrap();
 
                 Status::Continue
             })
         });
 
     let config = Config {
-        actions: Actions::REPLACE_BODY,
-        ..Default::default()
+        actions: Actions::ADD_HEADER | Actions::CHANGE_HEADER,
+        ..default_config()
     };
 
     let milter = Milter::spawn(LOCALHOST, callbacks, config).await.unwrap();
@@ -169,10 +108,84 @@ async fn replace_body_chunks() {
     let reply = client.read_reply().await.unwrap();
     assert!(matches!(
         reply,
-        Reply::OptNeg {
-            actions: Actions::REPLACE_BODY,
-            ..
+        Reply::OptNeg { actions, .. } if actions == Actions::ADD_HEADER | Actions::CHANGE_HEADER
+    ));
+
+    client
+        .write_command(Command::BodyEnd(Bytes::new()))
+        .await
+        .unwrap();
+
+    let reply = client.read_reply().await.unwrap();
+    assert_eq!(
+        reply,
+        Reply::AddHeader {
+            name: c_str!("name").into(),
+            value: c_str!("<value42>").into(),
         }
+    );
+
+    let reply = client.read_reply().await.unwrap();
+    assert_eq!(
+        reply,
+        Reply::ChangeHeader {
+            name: c_str!("other").into(),
+            index: 1,
+            value: c_str!("some value").into(),
+        }
+    );
+
+    let reply = client.read_reply().await.unwrap();
+    assert_eq!(reply, Reply::Continue);
+
+    client.write_command(Command::Quit).await.unwrap();
+
+    client.disconnect().await.unwrap();
+
+    milter.shutdown().await.unwrap();
+}
+
+#[tokio::test]
+async fn replace_body_chunks() {
+    init_tracing_subscriber();
+
+    let new_body_size = 2_000_000;
+
+    let callbacks = Callbacks::<()>::new()
+        .on_eom(move |cx| {
+            Box::pin(async move {
+                // New body with large size, greater than chunk size (an
+                // implementation detail).
+                let body = vec![b'x'; new_body_size];
+
+                cx.actions.replace_body(&body).await.unwrap();
+
+                Status::Continue
+            })
+        });
+
+    let config = Config {
+        actions: Actions::REPLACE_BODY,
+        ..default_config()
+    };
+
+    let milter = Milter::spawn(LOCALHOST, callbacks, config).await.unwrap();
+
+    let mut client = Client::connect(milter.addr()).await.unwrap();
+
+    client
+        .write_command(Command::OptNeg(OptNegPayload {
+            version: PROTOCOL_VERSION,
+            actions: Actions::all(),
+            opts: ProtoOpts::all(),
+        }))
+        .await
+        .unwrap();
+
+    let reply = client.read_reply().await.unwrap();
+    assert!(matches!(
+        reply,
+        Reply::OptNeg { actions, .. } if actions == Actions::REPLACE_BODY
     ));
 
     client
@@ -217,29 +230,28 @@ async fn replace_body_chunks() {
 }
 
 #[tokio::test]
-async fn header_actions() {
+async fn context_actions_error() {
     init_tracing_subscriber();
 
     let callbacks = Callbacks::<()>::new()
-        .on_eom(move |cx| {
-            Box::pin(async move {
-                // Also try sending a non-static string to `add_header` (to test
-                // `IntoCString + Send` parameter type).
-                let value = format!("<{}{}>", "value", 42);
+        .on_eom(|cx| {
+            Box::pin(async {
+                let result = cx.actions.add_recipient("").await;
+                assert!(matches!(result, Err(ActionError::InvalidParam)));
 
-                cx.actions.add_header("name", value.as_str()).await.unwrap();
-                cx.actions
-                    .change_header("other", 1, Some("some value"))
-                    .await
-                    .unwrap();
+                let result = cx
+                    .actions
+                    .add_recipient_ext("user@example.com", None::<CString>)
+                    .await;
+                assert!(matches!(result, Err(ActionError::NotAvailable)));
 
                 Status::Continue
             })
         });
 
     let config = Config {
-        actions: Actions::ADD_HEADER | Actions::CHANGE_HEADER,
-        ..Default::default()
+        actions: Actions::ADD_RCPT,
+        ..default_config()
     };
 
     let milter = Milter::spawn(LOCALHOST, callbacks, config).await.unwrap();
@@ -256,32 +268,16 @@ async fn header_actions() {
         .unwrap();
 
     let reply = client.read_reply().await.unwrap();
-    assert!(matches!(reply, Reply::OptNeg { .. }));
+    assert!(matches!(
+        reply,
+        Reply::OptNeg { actions, .. } if actions == Actions::ADD_RCPT
+    ));
 
     client
         .write_command(Command::BodyEnd(Bytes::new()))
         .await
         .unwrap();
 
-    let reply = client.read_reply().await.unwrap();
-    assert_eq!(
-        reply,
-        Reply::AddHeader {
-            name: c_str!("name").into(),
-            value: c_str!("<value42>").into(),
-        }
-    );
-
-    let reply = client.read_reply().await.unwrap();
-    assert_eq!(
-        reply,
-        Reply::ChangeHeader {
-            name: c_str!("other").into(),
-            index: 1,
-            value: c_str!("some value").into(),
-        }
-    );
-
     let reply = client.read_reply().await.unwrap();
     assert_eq!(reply, Reply::Continue);
 
