@@ -4,11 +4,11 @@ pub use crate::common::*;
 use byte_strings::c_str;
 use indymilter::{
     message::{
-        commands::{Command, ConnInfoPayload, MacroPayload, OptNegPayload},
-        replies::Reply,
+        command::{Command, ConnInfoPayload, MacroPayload, OptNegPayload},
+        reply::Reply,
         PROTOCOL_VERSION,
     },
-    Actions, Callbacks, ProtoOpts, SocketInfo, Stage, Status,
+    Actions, Callbacks, Config, ProtoOpts, SocketInfo, Stage, Status,
 };
 use std::{
     collections::HashMap,
@@ -17,25 +17,73 @@ use std::{
     sync::{Arc, Mutex},
 };
 
+#[tokio::test]
+async fn default_opts() {
+    init_tracing_subscriber();
+
+    let callbacks = Callbacks::<()>::new().on_eoh(|_| Box::pin(async { Status::Continue }));
+
+    let milter = Milter::spawn(LOCALHOST, callbacks, default_config())
+        .await
+        .unwrap();
+
+    let mut client = Client::connect(milter.addr()).await.unwrap();
+
+    client
+        .write_command(Command::OptNeg(OptNegPayload {
+            version: PROTOCOL_VERSION,
+            actions: Actions::all(),
+            opts: ProtoOpts::all(),
+        }))
+        .await
+        .unwrap();
+
+    let reply = client.read_reply().await.unwrap();
+    assert_eq!(
+        reply,
+        Reply::OptNeg {
+            version: PROTOCOL_VERSION,
+            actions: Actions::empty(),
+            opts: ProtoOpts::NO_CONNECT
+                | ProtoOpts::NO_HELO
+                | ProtoOpts::NO_MAIL
+                | ProtoOpts::NO_RCPT
+                | ProtoOpts::NO_DATA
+                | ProtoOpts::NO_HEADER
+                | ProtoOpts::NO_BODY
+                | ProtoOpts::NO_UNKNOWN,
+            macros: HashMap::new(),
+        }
+    );
+
+    client.write_command(Command::Quit).await.unwrap();
+
+    client.disconnect().await.unwrap();
+
+    milter.shutdown().await.unwrap();
+}
+
 #[tokio::test]
 async fn request_macros() {
     init_tracing_subscriber();
 
     let callbacks = Callbacks::<()>::new()
-        .on_negotiate(|cx, _actions, _opts| {
-            Box::pin(async move {
+        .on_negotiate(|cx, _, _| {
+            Box::pin(async {
                 cx.requested_actions = Actions::empty();
                 cx.requested_opts = ProtoOpts::empty();
-
-                cx.requested_macros.insert(Stage::Helo, c_str!("{what}").into());
+                cx.requested_macros = HashMap::from([(Stage::Helo, c_str!("{what}").into())]);
 
                 Status::Continue
             })
         });
 
-    let milter = Milter::spawn(LOCALHOST, callbacks, Default::default())
-        .await
-        .unwrap();
+    let config = Config {
+        actions: Actions::QUARANTINE,  // overridden in negotiate
+        ..default_config()
+    };
+
+    let milter = Milter::spawn(LOCALHOST, callbacks, config).await.unwrap();
 
     let mut client = Client::connect(milter.addr()).await.unwrap();
 
@@ -87,7 +135,7 @@ async fn out_of_order_macros() {
             })
     };
 
-    let milter = Milter::spawn(LOCALHOST, callbacks, Default::default())
+    let milter = Milter::spawn(LOCALHOST, callbacks, default_config())
         .await
         .unwrap();
 
@@ -143,15 +191,15 @@ async fn requested_actions_not_available() {
     init_tracing_subscriber();
 
     let callbacks = Callbacks::<()>::new()
-        .on_negotiate(|cx, _actions, _opts| {
-            Box::pin(async move {
+        .on_negotiate(|cx, _, _| {
+            Box::pin(async {
                 cx.requested_actions = Actions::ADD_RCPT | Actions::DELETE_RCPT;
 
                 Status::Continue
             })
         });
 
-    let milter = Milter::spawn(LOCALHOST, callbacks, Default::default())
+    let milter = Milter::spawn(LOCALHOST, callbacks, default_config())
         .await
         .unwrap();
 
